let {
  THREE,
  Edge,
  Node,
  OrbitControls,
  d3,
  ee,
  DEF_NODE_COLOR,
  DEF_NODE_HIGHLIGHT,
  DEF_ADJACENT_HIGHLIGHT,
  DEF_NODE_HTEXTURE,
  DEF_NODE_TEXTURE,
  DEF_NODE_NO_HTEX,
  DEF_NODE_NO_TEX,
  DEF_NODE_NO_HCOLOR,
  DEF_NODE_NO_COLOR,
  DEF_NODE_NO_HIGHLIGHT,
  DEF_NODE_SHAPE,
  DEF_NODE_HSHAPE,
  DEF_NODE_SIZE,
  DEF_LAYOUT,
  DEF_LINES,
  DEF_MAP,
  DEF_TEXT_SCALE,
  DEF_TEXT_SIZE,
  DEF_TEXT_X_OFFSET,
  DEF_TEXT_Y_OFFSET,
  DEF_TEXT_BACK_COLOR,
  DEF_TEXT_OPACITY,
  TEXT_WIREFRAME
} = require('./imports').default;

var Utils = require('./utils/utils');

var Frame = function (graph, options) {
  // Needed to reference "this" in functions
  var self = this;

  // Options
  this.inGraph = graph;
  this.options = options || {
    "layout": DEF_LAYOUT
  };
  this.layout_options = this.options.graphLayout || {};

  this.lineObjects = [];
  this.lineIndices = [];
  this.MAX_LINES = 100000;
  this.textureLoader = new THREE.TextureLoader();
  this.textureLoader.crossOrigin = true;
  this.oldIntersect = -1;
  this.selection = [];
  this.fov = 60;
  this.near = 1;
  this.far = 5000;
  this.mouseStart = new THREE.Vector3(0, 0, 0);
  this.mouseEnd = new THREE.Vector3(0, 0, 0);
  this.dragging = null;
  this.showBox = false;
  this.renderWidth = 500;
  this.renderHeight = 500;
  this.maxZoom = 1000;
  this.paused = false;
  this.mouseDown = false;
  this.ee = ee({});
  this.layoutInit = true;

  require('./setup')(this);
  require('./addremove')(this);
  require('./select')(this);
  require('./modify')(this);
  require('./draw')(this);
  require('./update')(this);

  /**
   *  Starting point, run once to create scene
   */
  this.display = function () {
    if (graph == null) {
      Utils.randomGraph(self);
    }
    this.animate();
  }

  /**
   *  Creates loop called on every animation frame
   */
  let fps = 48;
  let now;
  let then = Date.now();
  let interval = 1000/fps;
  let delta;
  this.animate = function () {
    requestAnimationFrame( self.animate );

    now = Date.now();
    delta = now - then;
    if (delta > interval) {
      then = now - (delta % interval);
      self.controls.update();
      self.render();
    }
  };

  /**
   *  Set initial properties
   */
  this.init = function () {
    self.renderer = new THREE.WebGLRenderer({
      alpha: true
    });
    self.cssRenderer = new THREE.CSS3DRenderer();
    self.element = document.getElementById('graph-container');
    self.clientRect = self.element.getBoundingClientRect();
    self.width = self.clientRect.width;
    self.height = self.clientRect.height;
    self.aspect = self.width / self.height;
    self.resolution = new THREE.Vector2(self.width, self.height);
    self.renderer.setSize(self.width, self.height);
    self.cssRenderer.setSize(self.width, self.height);
    self.cssRenderer.domElement.style.position = 'absolute';
    self.cssRenderer.domElement.style.top = 0;
    self.renderer.setPixelRatio(window.devicePixelRatio);

    self.setupCamera();
    self.setupGeometry();
    self.setupSelect();

    self.element.appendChild(self.renderer.domElement);
    self.element.appendChild(self.cssRenderer.domElement);

    self.canvas = document.querySelector('graph-container');

    self.setupLayout();

    self.setupGraph();

    window.addEventListener('resize', self.onWindowResize, false);
  };

  /**
   *  Draws graphics
   */
  this.render = function () {
    self.updateCamera();
    self.updateNodes();
    self.updateEdges();
    if (self.options.layout == "d3") {
      if (self.layoutInit == true) {
        var nodes = []
        self.graph.forEachNode(function (node) {
          nodes.push(node);
        });
        self.force.nodes(nodes);
        // var simNodes = self.force.nodes();
        // var idToIndex = {};
        // for (var i = 0; i < simNodes.length; i++) {
        //   idToIndex[simNodes[i].id] = simNodes[i].index;
        // }
        // for (var i = 0; i < self.lineIndices.length; i++) {
        //   self.lineIndices[i].source = idToIndex[self.lineIndices[i].source];
        //   self.lineIndices[i].target = idToIndex[self.lineIndices[i].target];
        // }
        self.force.force("link", d3.forceLink(self.lineIndices));
        self.force.restart();
        self.force.stop();
        for (var i = 0, n = Math.ceil(Math.log(self.force.alphaMin()) / Math.log(1 - self.force.alphaDecay())); i < n; ++i) {
          self.force.tick();
        }
        self.layoutInit = false;
      }
    }
    self.renderer.setViewport(0, 0, 1 * self.width, 1 * self.height);
    self.renderer.setScissor(0, 0, 1 * self.width, 1 * self.height);
    self.renderer.setScissorTest(true);
    self.renderer.render(self.scene, self.ccamera);
    self.cssRenderer.render(self.scene, self.ccamera);
    if (DEF_MAP == "yes") {
      self.renderer.setViewport(0, 0, 0.2 * (self.width / self.aspect), 0.2 * (self.height / self.aspect));
      self.renderer.setScissor(0, 0, 0.2 * (self.width / self.aspect), 0.2 * (self.height / self.aspect));
      self.renderer.setScissorTest(true);
      self.renderer.render(self.scene, self.ccamera2);
    }
  };

  this.updateGraph = function (graph) {
    graph.forEachNode(function (node) {
      var oldNode = self.graph.getNode(node.id);
      if (!oldNode) {
        self.addNode(node);
      } else {
        self.updateNode(oldNode, node);
      }
    });
    graph.forEachLink(function (link) {
      if (!self.graph.getLink(link.fromId, link.toId)) {
        self.addEdge(self.graph.getNode(link.fromId), self.graph.getNode(link.toId));
      }
    });

    self.graph.forEachNode(function (node) {
      var oldNode = graph.getNode(node.id);
      if (!oldNode) {
        self.removeNode(node);
      }
    });

    self.force.alpha(1);
    self.force.restart();
    self.force.stop();
  }

  this.onSelectNodes = function () {
    for (var i = 0; i < self.selection.length; i++) {
      if (!self.selection[i].pinned) {
        self.selection[i].fx = null;
        self.selection[i].fy = null;
      }
    }
  };

  this.pauseLayout = () => {
    self.paused = true;

  };
   window.pause = this.pauseLayout;

  this.resumeLayout = () => {
    self.paused = false;
  };

  this.mapSelectedNodes = () => {
    self.mapNodeAttributes(["degree", "log", "scalexy"], self.selection);
  };

  this.mapAllNodes = () => {
    self.mapNodeAttributes(["degree", "log", "scalexy"]);
  };

  this.pinSelectedNodes = () => {
    self.mapNodeAttributes([true, "", "pinned"], self.selection);
  };

  this.unpinSelectedNodes = () => {
    self.mapNodeAttributes([false, "", "pinned"], self.selection);
  };

  // Size in em, not px
  this.setLabelFontSize = (size) => {
    self.graph.forEachNode(function (node) {
      self.changeLabelFontSize(node, size);
    });
  }

  this.setLabelLength = (numChars) => {
    self.graph.forEachNode(function (node) {
      self.changeLabelLength(node, numChars);
    })
  }

  this.updateNode = function (node, newNode) {
    if (newNode.data) {
      if (node.data.color) {
        node.data.color = newNode.data.color;
        node.renderData.color = node.data.color;
        node.renderData.draw_object.material.color.set(new THREE.Color(node.data.color));
      }
      if (newNode.data.size && newNode.data.size > 0) {
        node.data.size = newNode.data.size;
        node.renderData.size = newNode.data.size;
        node.renderData.draw_object.scale.set(newNode.data.size, newNode.data.size, 1);
      }
      if (newNode.data.label) {
        node.data.label = newNode.data.label;
        node.renderData.label = node.data.label;
        node.renderData.textHolder.children[0].element.childNodes[0].innerText = node.data.label;
      }
    }
  }

  this.setCanvasSize = function(size) {
    self.setBoundarySize(size);
  }
};

exports.Frame = Frame;
